/*
 * Tapidbus library
 * Copyright (C) 2005 INdT
 * @author Luiz Augusto von Dentz <luiz.dentz@indt.org.br>
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA02111-1307USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "tapidbus.h"
#include "tapidbus_compat.h"

#include <dbus/dbus-glib-lowlevel.h>

typedef struct _TapiDBusPendingCallData TapiDBusPendingCallData;
struct _TapiDBusPendingCallData {
    TapiDBus *td;
    TapiDBusPendingCallNotifyFunction *function;
    gpointer user_data;
};

static DBusConnection    *_dbus_connection_new         (void);
static gboolean           _dbus_connection_setup       (DBusConnection  *conn,
                                                        const gchar     *service_name,
                                                        const gchar     *path,
                                                        TapiDBusMethod  *methods,
                                                        TapiDBusSignal  *signals,
                                                        gpointer         filter_data);

static DBusHandlerResult  _dbus_filter_cb              (DBusConnection  *conn,
                                                        DBusMessage     *message,
                                                        void            *user_data);

static DBusMessage       *_dbus_build_method_call      (const gchar     *destination,
                                                        const gchar     *path,
                                                        const gchar     *iface,
                                                        const gchar     *method,
                                                        gint             type,
                                                        va_list         *data);

static void               _dbus_pending_call_notify_cb (DBusPendingCall *pending,
                                                        void            *user_data);

static const DBusObjectPathVTable vtable = { NULL, &_dbus_filter_cb };

/**
 * Create a new TapiDBus.
 *
 * @return a new TapiDBus or NULL, if some error occurred
 */
TapiDBus *
tapidbus_new (void)
{
    TapiDBus *td;

    td = g_new0 (TapiDBus, 1);
    td->service_name = NULL;
    td->path = NULL;
    td->conn = _dbus_connection_new ();
    if (!td->conn) {
        g_free (td);
        return NULL;
    }
    
    return td;    
}

/**
 * Create a new TapiDBus.
 *
 * @param service_name the service name to be initialized
 * @param path the path name
 * @param methods the struct that contains the methods provided by the given interface
 * @param signals the struct that contains the signals provided by the given interface
 *
 * @return a new TapiDBus or NULL, if some error occurred
 */
TapiDBus *
tapidbus_new_with_service (const gchar    *service_name,
                           const gchar    *path,
                           TapiDBusMethod *methods,
                           TapiDBusSignal *signals)
{
    TapiDBus *td;

    g_return_val_if_fail (path != NULL, NULL);
    g_return_val_if_fail (methods != NULL || signals != NULL, NULL);

    if (methods && !service_name) {
        g_warning ("%s (%d): error creating TapiDBus object. You need a service_name in order to use methods\n",
                __FUNCTION__, __LINE__);
        return NULL;
    }
    
    td = g_new0 (TapiDBus, 1);
    g_assert (td);

    if (service_name)
        td->service_name = g_strdup (service_name);
    td->path = g_strdup (path);
    td->methods = methods;
    td->signals = signals;

    td->conn = _dbus_connection_new ();
    if (td->conn) {
        if (_dbus_connection_setup (td->conn,
                    service_name,
                    path,
                    methods, 
                    signals,
                    td)) {
            return td;
        }
    }

    if (td->service_name)
        g_free (td->service_name);
    g_free (td->path);
    g_free (td);

    return NULL;
}

/**
 * Set the service for TapiDBus. If some error occurs
 * the TapiDBus* will be invalidated.
 *
 * @param td Pointer to the TapiDBus object to be modified
 * @param service_name the interface name to be initialized
 * @param path the path name
 * @param methods the struct that contains the methods provided by the given interface
 * @param signals the struct that contains the signals provided by the given interface
 *
 * @return TRUE on success, otherwise FALSE.
 */
gboolean
tapidbus_set_service (TapiDBus       *td,
                      const gchar    *service_name,
                      const gchar    *path,
                      TapiDBusMethod *methods,
                      TapiDBusSignal *signals)
{
    g_return_val_if_fail (td != NULL, FALSE);
    g_return_val_if_fail (td->service_name == NULL, FALSE);
    g_return_val_if_fail (td->path == NULL, FALSE);
    g_return_val_if_fail (path != NULL, FALSE);
    g_return_val_if_fail (methods != NULL || signals != NULL, FALSE);
   
    if (methods && !service_name) {
        g_warning ("%s (%d): error creating TapiDBus object. You need a service_name in order to use methods\n",
                __FUNCTION__, __LINE__);
        return FALSE;
    }

    if (!td->conn) {
        td->conn = _dbus_connection_new ();
        if (!td->conn) {
            return FALSE;
        }
    }

    if (service_name)
        td->service_name = g_strdup (service_name);
    td->path = g_strdup (path);
    td->methods = methods;
    td->signals = signals;

    if (!_dbus_connection_setup (td->conn,
                service_name,
                path,
                methods, 
                signals,
                td)) {
        td->conn = NULL;
        if (td->service_name) {
            g_free (td->service_name); 
            td->service_name = NULL;
        }
        g_free (td->path); 
        td->path = NULL;
        return FALSE;
    }

    return TRUE;
}

/**
 * Stop dbus interface
 *
 * @param td the TapiDBus to be deinitialized
 */
void
tapidbus_destroy (TapiDBus *td)
{
    g_return_if_fail (td != NULL);
  
    if (td->conn) {
        DBusError error;
        guint i;
        gchar *match;
        
        dbus_error_init (&error);

        if (td->methods) {
            for (i = 0; td->methods[i].iface; ++i) {
                match = g_strdup_printf ("type='method_call',path='%s',interface='%s',member='%s'", 
                        td->path, td->methods[i].iface, td->methods[i].name);
                dbus_bus_remove_match (td->conn, match, &error);
                // g_print ("%s (%d): removing match \"%s\"\n",
                //         __FUNCTION__, __LINE__, match);
                if (dbus_error_is_set (&error)) {
                    g_warning ("%s (%d): error removing match \"%s\" - %s",
                        __FUNCTION__, __LINE__, match, error.message);
                    dbus_error_free (&error);
                    dbus_error_init (&error);
                }
                g_free (match);
            }
        }
        
        if (td->signals) {
            for (i = 0; td->signals[i].iface; ++i) {
                match = g_strdup_printf ("type='signal',path='%s',interface='%s',member='%s'", 
                        td->path, td->signals[i].iface, td->signals[i].name);
                dbus_bus_remove_match (td->conn, match, &error);
                // g_print ("%s (%d): removing match \"%s\"\n",
                //         __FUNCTION__, __LINE__, match);
                if (dbus_error_is_set (&error)) {
                    g_warning ("%s (%d): error removing match \"%s\" - %s",
                        __FUNCTION__, __LINE__, match, error.message);
                    dbus_error_free (&error);
                    dbus_error_init (&error);
                }
                g_free (match);
            }
        }
      
        if (td->path)
            dbus_connection_unregister_object_path (td->conn,
                td->path); 

        // dbus_connection_unref (td->conn);
        td->conn = NULL;
    }

    if (td->service_name) {
        g_free (td->service_name);
        td->service_name = NULL;
    }

    if (td->path) {
        g_free (td->path);
        td->path = NULL;
    }

    g_free (td);
}

/**
 * Send a message over DBUS.
 *
 * @param td the TapiDBus that will be used to send the message
 * @param destination the message destination 
 * @param path the path to access the interface
 * @param iface the interface name
 * @param method the method name
 * @param type type of the first argument
 * @param ... value of first argument, list of additional type-value pairs
 *
 * @return TRUE if message could be sent.
 */
gboolean
tapidbus_send (TapiDBus    *td,
               const gchar *destination,
               const gchar *path,
               const gchar *iface,
               const gchar *method,
               gint         type, 
               ...)
{
    va_list arg_list;
    DBusMessage *message;
    gboolean ret = TRUE;

    g_return_val_if_fail (td != NULL, FALSE);
    g_return_val_if_fail (td->conn != NULL, FALSE);
    g_return_val_if_fail (path != NULL, FALSE);
    g_return_val_if_fail (iface != NULL, FALSE);
    g_return_val_if_fail (method != NULL, FALSE);

    va_start (arg_list, type);

    message = _dbus_build_method_call (destination, path, 
            iface, method, type, &arg_list);
    dbus_message_set_no_reply (message, TRUE);
    if (!dbus_connection_send (td->conn, message, NULL))
        ret = FALSE;
    dbus_connection_flush (td->conn);
    dbus_message_unref (message);

    va_end (arg_list);

    return ret;
}

/**
 * Send a message over DBUS with pending reply. A callback must be given to handle the reply.
 *
 * @param td the TapiDBus that will be used to send the message
 * @param destination the message destination 
 * @param path the path to access the interface
 * @param iface the interface name
 * @param method the method name
 * @param function Callback function
 * @param user_data User data to be sent to callback
 * @param timeout_milliseconds Timeout of the message
 * @param type type of the first argument
 * @param ... value of first argument, list of additional type-value pairs
 *
 * @return TRUE if message could be sent.
 */
DBusPendingCall *
tapidbus_send_with_reply (TapiDBus                          *td,
                          const gchar                       *destination,
                          const gchar                       *path,
                          const gchar                       *iface,
                          const gchar                       *method,
                          TapiDBusPendingCallNotifyFunction  function,
                          gpointer                           user_data,
                          gint                               timeout_milliseconds,
                          gint                               type,
                          ...)
{
    va_list arg_list;
    DBusMessage *message;
    DBusPendingCall *pending = NULL;

    g_return_val_if_fail (td != NULL, FALSE);
    g_return_val_if_fail (td->conn != NULL, FALSE);
    g_return_val_if_fail (path != NULL, FALSE);
    g_return_val_if_fail (iface != NULL, FALSE);
    g_return_val_if_fail (method != NULL, FALSE);

    va_start (arg_list, type);

    message = _dbus_build_method_call (destination, path, 
            iface, method, type, &arg_list);
    if (dbus_connection_send_with_reply (td->conn, message,
                &pending, timeout_milliseconds)) {
        if (function) {
            TapiDBusPendingCallData *data = g_new0 (TapiDBusPendingCallData, 1);
            data->td = td;
            data->function = function;
            data->user_data = user_data;
            if (!dbus_pending_call_set_notify (pending,
                    _dbus_pending_call_notify_cb, 
                    data,
                    NULL)) {
                g_free (data);
                pending = NULL;
            }
        }
    }
    else {
        pending = NULL;
    }
    dbus_message_unref (message);

    va_end(arg_list);

    return pending;
}

/**
 * Send a message over DBUS with blocking reply.
 *
 * @param td the TapiDBus that will be used to send the message
 * @param destination the message destination 
 * @param path the path to access the interface
 * @param iface the interface name
 * @param method the method name
 * @param timeout_milliseconds Timeout of the message
 * @param type type of the first argument
 * @param ... value of first argument, list of additional type-value pairs
 *
 * @return TRUE if message could be sent.
 */
DBusMessage *
tapidbus_send_with_reply_block (TapiDBus    *td,
                                const gchar *destination,
                                const gchar *path,
                                const gchar *iface,
                                const gchar *method,
                                gint         timeout_milliseconds,
                                gint         type, 
                                ...)
{
    va_list arg_list;
    DBusMessage *message, *reply;
    DBusError error;

    g_return_val_if_fail (td != NULL, NULL);
    g_return_val_if_fail (td->conn != NULL, NULL);
    g_return_val_if_fail (path != NULL, NULL);
    g_return_val_if_fail (iface != NULL, NULL);
    g_return_val_if_fail (method != NULL, NULL);

    va_start (arg_list, type);

    dbus_error_init (&error);
    message = _dbus_build_method_call (destination, path, 
            iface, method, type, &arg_list);
    reply = dbus_connection_send_with_reply_and_block (td->conn,
            message, timeout_milliseconds, &error);
    if (dbus_error_is_set (&error)) {
        dbus_error_free (&error);
    }

    dbus_message_unref (message);

    va_end (arg_list);

    return reply;
}

/**
 * Emit a signal over DBUS.
 *
 * @param td the TapiDBus that will be used to send the message
 * @param path the path to access the interface
 * @param iface the interface name
 * @param method the method name
 * @param type type of the first argument
 * @param ... value of first argument, list of additional type-value pairs
 *
 * @return TRUE if message could be sent.
 */
gboolean
tapidbus_emit_signal (TapiDBus    *td,
                      const gchar *path,
                      const gchar *iface,
                      const gchar *method,
                      gint         type, 
                      ...)
{
    va_list arg_list;
    DBusMessage *message;
    gboolean ret = TRUE;

    g_return_val_if_fail (td != NULL, FALSE);
    g_return_val_if_fail (td->conn != NULL, FALSE);
    g_return_val_if_fail (path != NULL, FALSE);
    g_return_val_if_fail (iface != NULL, FALSE);
    g_return_val_if_fail (method != NULL, FALSE);

    va_start (arg_list, type);

    /* Create a new message to send */
    message = dbus_message_new_signal (path,
            iface,
            method);
    g_assert (message);

    /* Check and append all args */
    dbus_message_append_args_valist (message, type, arg_list);
   
    if (!dbus_connection_send (td->conn, message, NULL))
        ret = FALSE;
    dbus_connection_flush (td->conn);
    dbus_message_unref (message);

    va_end (arg_list);

    return ret;
}

/**
 * Build a reply message.
 * 
 * @param request Request message
 * @param type type of the first argument
 * @param ... value of first argument, list of additional type-value pairs
 * 
 * @return Reply message
 */
DBusMessage *
tapidbus_build_method_return (DBusMessage *request,
                              gint         type,
                              ...)
{
    va_list arg_list;
    DBusMessage *reply;
    
    g_return_val_if_fail (request != NULL, NULL);

    va_start (arg_list, type);

    reply = dbus_message_new_method_return (request);
    if (!reply)
        return NULL;
            
    /* Check and append all args */
    dbus_message_append_args_valist (reply, type, arg_list);

    va_end (arg_list);
    
    return reply;
}

/**
 * Get one integer (32 bytes) value from a reply message.
 * 
 * @param message Reply message from a request
 * @param value Pointer to a variable that will store the value
 * 
 * @return Returns TRUE on success, otherwise FALSE.
 */
gboolean
tapidbus_get_int32 (DBusMessage *message, 
                    gint        *value)
{
    gboolean ret = TRUE;
    DBusError error;

    g_return_val_if_fail (message != NULL, FALSE);
 
    dbus_error_init (&error);
    if (!dbus_message_get_args (message, &error,
                DBUS_TYPE_INT32, value,
                DBUS_TYPE_INVALID)) {
        ret = FALSE;
    }
    dbus_error_free (&error);

    return ret;
}

/**
 * Get one string value from a reply message.
 * 
 * @param message Reply message from a request
 * @param value Pointer to a variable that will store the value
 * 
 * @return TRUE on success, otherwise FALSE.
 */
gboolean
tapidbus_get_string (DBusMessage  *message, 
                     gchar       **value)
{
    gboolean ret = TRUE;
    DBusError error;

    g_return_val_if_fail (message != NULL, FALSE);
 
    dbus_error_init (&error);
    if (!dbus_message_get_args (message, &error,
                DBUS_TYPE_STRING, value,
                DBUS_TYPE_INVALID)) {
        ret = FALSE;
    }
    dbus_error_free (&error);

    return ret;
}

/**
 * Get one boolean value from a reply message.
 * 
 * @param message Reply message from a request
 * @param value Pointer to a variable that will store the value
 * 
 * @return Returns TRUE on success, otherwise FALSE.
 */
gboolean
tapidbus_get_boolean (DBusMessage *message, 
                      gboolean    *value)
{
    gboolean ret = TRUE;
    DBusError error;

    g_return_val_if_fail (message != NULL, FALSE);

    dbus_error_init (&error);
    if (!dbus_message_get_args (message, &error,
                DBUS_TYPE_BOOLEAN, value,
                DBUS_TYPE_INVALID)) {
        ret = FALSE;
    }
    dbus_error_free (&error);

    return ret;
}

/**
 * Get one double value from a reply message.
 * 
 * @param message Reply message from a request
 * @param value Pointer to a variable that will store the value
 * 
 * @return Returns TRUE on success, otherwise FALSE.
 */
gboolean
tapidbus_get_double (DBusMessage *message, 
                     gdouble     *value)
{
    gdouble ret = TRUE;
    DBusError error;

    g_return_val_if_fail (message != NULL, FALSE);

    dbus_error_init (&error);
    if (!dbus_message_get_args (message, &error,
                DBUS_TYPE_DOUBLE, value,
                DBUS_TYPE_INVALID)) {
        ret = FALSE;
    }
    dbus_error_free (&error);

    return ret;
}

/**
 * Get all arguments from a request message.
 * 
 * @param message Request message from a request
 * @param type type of the first argument
 * @param ... value of first argument, list of additional type-value pairs
 * 
 * @return Returns FALSE if an error occurs otherwise TRUE
 */
gboolean
tapidbus_get_args (DBusMessage *message,
                   gint         type,
                   ...)
{
    DBusError error;
    va_list arg_list;
    gboolean ret = TRUE;

    g_return_val_if_fail (message != NULL, FALSE);

    va_start (arg_list, type);
    dbus_error_init (&error);
    if (!dbus_message_get_args_valist (message, &error,
                type, arg_list)) {
        ret = FALSE;
    }
    dbus_error_free (&error);
    va_end (arg_list);

    return ret;   
}    

static DBusConnection *
_dbus_connection_new (void)
{
    DBusConnection *conn;
    DBusError error;

    dbus_connection_set_change_sigpipe (TRUE);

    dbus_error_init (&error);
    conn = dbus_bus_get (DBUS_BUS_SESSION, &error);
    if (conn == NULL) {
        g_warning ("%s (%d): cannot open connection to message bus: %s\n",
                __FUNCTION__, __LINE__, error.message);
        dbus_error_free (&error);
        return NULL;
    }

    return conn;
}

static gboolean
_dbus_connection_setup (DBusConnection *conn,
                        const gchar    *service_name,
                        const gchar    *path,
                        TapiDBusMethod *methods,
                        TapiDBusSignal *signals,
                        gpointer        filter_data)
{
    DBusError error;
    guint i;
    gchar *match;

    dbus_connection_setup_with_g_main (conn, NULL);
    dbus_connection_set_exit_on_disconnect (conn, FALSE);
        
    dbus_error_init (&error);

    if (service_name)
#if CHECK_DBUS_VERSION(0, 30)
        dbus_bus_request_name (conn, service_name, 0, &error);
#else
        dbus_bus_acquire_service (conn, service_name, 0, &error);
#endif

    if (dbus_error_is_set (&error)) {
        g_warning ("%s (%d): failed to assign a name to dbus connection %s",
            __FUNCTION__, __LINE__, error.message);
        dbus_error_free (&error);
        // dbus_connection_unref (conn);
        conn = NULL;
        return FALSE;
    }
   
    if (!dbus_connection_register_object_path (conn,
            path, 
            &vtable, 
            filter_data)) {
        // dbus_connection_unref (conn);
        conn = NULL;
        return FALSE;
    }

    if (methods) {
        for (i = 0; methods[i].iface; ++i) {
            match = g_strdup_printf ("type='method_call',path='%s',interface='%s',member='%s'", 
                    path, methods[i].iface, methods[i].name);
            dbus_bus_add_match (conn, match, &error);
            // g_print ("%s (%d): adding match \"%s\"\n",
            //         __FUNCTION__, __LINE__, match);
            if (dbus_error_is_set (&error)) {
                g_warning ("%s (%d): error adding match \"%s\" - %s",
                    __FUNCTION__, __LINE__, match, error.message);
                dbus_error_free (&error);
                dbus_error_init (&error);
            }
            g_free (match);
        }
    }
    
    if (signals) {
        for (i = 0; signals[i].iface; ++i) {
            match = g_strdup_printf ("type='signal',path='%s',interface='%s',member='%s'", 
                    path, signals[i].iface, signals[i].name);
            dbus_bus_add_match (conn, match, &error);
            // g_print ("%s (%d): adding match \"%s\"\n",
            //         __FUNCTION__, __LINE__, match);
            if (dbus_error_is_set (&error)) {
                g_warning ("%s (%d): error adding match \"%s\" - %s",
                    __FUNCTION__, __LINE__, match, error.message);
                dbus_error_free (&error);
                dbus_error_init (&error);
            }
            g_free (match);
        }
    }
                 
    return TRUE;
}

static DBusHandlerResult
_dbus_filter_cb (DBusConnection *conn,
                 DBusMessage    *message,
                 void           *user_data)
{
    TapiDBus *td = (TapiDBus *) user_data;
    TapiDBusMethod *method;
    TapiDBusSignal *signal;
    DBusMessage *reply = NULL;
    gint type;
    const gchar *iface, *member, *signature;
    guint i;
    
    iface = dbus_message_get_interface (message);
    member = dbus_message_get_member (message);
    signature = dbus_message_get_signature (message);

    if (!iface || !member || !signature)
        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
    
    type = dbus_message_get_type (message);
    if (type == DBUS_MESSAGE_TYPE_METHOD_CALL &&
        td->methods) {
        for (i = 0; td->methods[i].iface; ++i) {
            method = &td->methods[i];
            if ((g_strcasecmp (iface, method->iface) == 0) &&
                (g_strcasecmp (member, method->name) == 0) &&
                (g_strcasecmp (signature, method->signature) == 0)) {
                reply = (method->handle_func) (td, message, method->user_data);
                if (reply) {
                    dbus_connection_send (td->conn, reply, NULL);
                    dbus_message_unref (reply);
                }
                return DBUS_HANDLER_RESULT_HANDLED;
            }
        }
    }
    else if (type == DBUS_MESSAGE_TYPE_SIGNAL &&
             td->signals) {
        for (i = 0; td->signals[i].iface; ++i) {
            signal = &td->signals[i];
            if ((g_strcasecmp (iface, signal->iface) == 0) &&
                (g_strcasecmp (member, signal->name) == 0) &&
                (g_strcasecmp (signature, signal->signature) == 0)) {
                (signal->handle_func) (td, message, signal->user_data);
                return DBUS_HANDLER_RESULT_HANDLED;
            }
        }       
    }

    return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

static DBusMessage *
_dbus_build_method_call (const gchar *destination,
                         const gchar *path,
                         const gchar *iface,
                         const gchar *method,
                         gint         type,
                         va_list     *data)
{
    DBusMessage *message;

    /* Create a new message to send */
    message = dbus_message_new_method_call (destination,
            path,
            iface,
            method);
    g_assert (message);

    /* Check and append all args */
    dbus_message_append_args_valist (message, type, *data);

    return message;
}

static void
_dbus_pending_call_notify_cb (DBusPendingCall *pending,
                              void            *user_data)
{
    TapiDBusPendingCallData *data = (TapiDBusPendingCallData *) user_data;
    DBusMessage *reply = NULL;

    g_return_if_fail (pending != NULL);
    g_return_if_fail (user_data != NULL);
   
    dbus_pending_call_ref (pending);

    // timeout
    if (dbus_pending_call_get_completed (pending))
        reply = dbus_pending_call_steal_reply (pending);
 
    (data->function) (data->td, reply, data->user_data);
   
    if (reply)
        dbus_message_unref (reply);

    dbus_pending_call_unref (pending);
    g_free (data);
}

